@@ -50,8 +50,12 @@ sub new {
 				    error_format => 'html',
 				    %p);
 
-    $self->interp->out_method(\$self->{output})
-        unless exists $p{out_method};
+    # If the user passed an out_method parameter, then we don't want
+    # to overwrite it later.  This doesn't handle the case where the
+    # user creates their own Interp object with a custom out_method
+    # and passes that to this method, though (grrr).
+    $self->{has_custom_out_method} = $p{out_method} ? 1 : 0;
+
     $self->interp->compiler->add_allowed_globals('$r');
     
     return $self;
@@ -80,25 +84,55 @@ sub _handler {
     my $r = $self->create_delayed_object('cgi_request', cgi => $p->{cgi});
     $self->interp->set_global('$r', $r);
 
-    $self->{output} = '';
+    # hack for testing
+    if (@_) {
+        $self->{output} = '';
+        $self->interp->out_method( \$self->{output} );
+    } elsif (! $self->{has_custom_out_method}) {
+        my $sent_headers = 0;
+
+        my $out_method = sub {
+
+            # Send headers if they have not been sent by us or by user.
+            # We use instance here because if we store $request we get a
+            # circular reference and a big memory leak.
+            if (!$sent_headers and HTML::Mason::Request->instance->auto_send_headers) {
+                unless ($r->http_header_sent) {
+                    $r->send_http_header();
+                }
+                $sent_headers = 1;
+            }
+
+            # We could perhaps install a new, faster out_method here that
+            # wouldn't have to keep checking whether headers have been
+            # sent and what the $r->method is.  That would require
+            # additions to the Request interface, though.
+
+            print STDOUT grep {defined} @_;
+        };
+
+        $self->interp->out_method($out_method);
+    }
 
     $self->interp->delayed_object_params('request', cgi_request => $r);
 
     my %args = $self->request_args($r);
 
     eval { $self->interp->exec($p->{comp}, %args) };
+
     if (my $err = $@) {
-        rethrow_exception($err)
-          unless isa_mason_exception($err, 'Abort')
-          or isa_mason_exception($err, 'Decline');
+
+        unless ( isa_mason_exception($err, 'Abort')
+                 or isa_mason_exception($err, 'Decline') ) {
+
+            rethrow_exception($err);
+        }
     }
 
     if (@_) {
-	# This is a secret feature, and should stay secret (or go away) because it's just a hack for the test suite.
+	# This is a secret feature, and should stay secret (or go
+	# away) because it's just a hack for the test suite.
 	$_[0] .= $r->http_header . $self->{output};
-    } else {
-        $r->send_http_header;
-	print $self->{output};
     }
 }
 
@@ -109,23 +143,57 @@ sub request_args {
     return $r->params;
 }
 
+
 ###########################################################
 package HTML::Mason::Request::CGI;
 # Subclass for HTML::Mason::Request object $m
 
+use HTML::Mason::Exceptions;
 use HTML::Mason::Request;
 use base qw(HTML::Mason::Request);
 
-use HTML::Mason::MethodMaker
-    ( read_only => [ 'cgi_request' ] );
+use Params::Validate qw(BOOLEAN);
+Params::Validate::validation_options( on_fail => sub { param_error( join '', @_ ) } );
 
-__PACKAGE__->valid_params( cgi_request => {isa => 'HTML::Mason::FakeApache'} );
+__PACKAGE__->valid_params
+    ( cgi_request => { isa => 'HTML::Mason::FakeApache' },
+
+      auto_send_headers => { parse => 'boolean', type => BOOLEAN, default => 1,
+                             descr => "Whether HTTP headers should be auto-generated" },
+    );
+
+use HTML::Mason::MethodMaker
+    ( read_only  => [ 'cgi_request' ],
+      read_write => [ 'auto_send_headers' ] );
 
 sub cgi_object {
     my $self = shift;
     return $self->{cgi_request}->query(@_);
 }
 
+#
+# Override this method to send HTTP headers if necessary.
+#
+sub exec
+{
+    my $self = shift;
+    my $r = $self->cgi_request;
+    my $retval;
+
+    eval { $retval = $self->SUPER::exec(@_) };
+
+    # On a success code, send headers if they have not been sent and
+    # if we are the top-level request. Since the out_method sends
+    # headers, this will typically only apply after $m->abort.
+    # On an error code, leave it to Apache to send the headers.
+    if (!$self->is_subrequest
+	and $self->auto_send_headers
+	and !$r->http_header_sent
+	and (!$retval or $retval==200)) {
+	$r->send_http_header();
+    }
+}
+
 sub redirect {
     my $self = shift;
     my $url = shift;
